#pragma once

#include <array>
#include <vector>
#include <cmath>
#include <iostream>

/** \brief Computes weights and integration points for the Gauss-Jacobi Quadrature
           \f$ \int_{-1}^1 f(x) (1-x)^\alpha (1+x)^\beta dx \f$.
           Transformed from [-1,1] to [0,1].
 *
 * \param[in] n Order of integration rule
 * \param[in] alf Exponent alpha. Must be > -1
 * \param[in] beta Exponent beta. Must be > -1
 * \param[out] xi vector of integration points
 * \param[out] wi vector of integration weights
 * \return
 *
 */
template<typename T>
inline void computeGaussJacobiRule(const int n,
                                   const T alf, const T bet,
                                   std::vector<T> & xi,
                                   std::vector<T> & wi)
{
    constexpr double  EPS = 3.0e-14;
    constexpr int  MAXIT = 10;
    xi.resize(n);
    wi.resize(n);

    int i,its,j;
    T alfbet,an,bn,r1,r2,r3;
    T a,b,c,p1,p2,p3,pp,temp,z=T(0),z1;
    for (i=1; i<=n; i++)
    {
        if (i == 1)
        {
            an=alf/n;
            bn=bet/n;
            r1=(1.0+alf)*(2.78/(4.0+n*n)+0.768*an/n);
            r2=1.0+1.48*an+0.96*bn+0.452*an*an+0.83*an*bn;
            z=1.0-r1/r2;
        }
        else if (i == 2)
        {
            r1=(4.1+alf)/((1.0+alf)*(1.0+0.156*alf));
            r2=1.0+0.06*(n-8.0)*(1.0+0.12*alf)/n;
            r3=1.0+0.012*bet*(1.0+0.25*std::abs(alf))/n;
            z -= (1.0-z)*r1*r2*r3;
        }
        else if (i == 3)
        {
            r1=(1.67+0.28*alf)/(1.0+0.37*alf);
            r2=1.0+0.22*(n-8.0)/n;
            r3=1.0+8.0*bet/((6.28+bet)*n*n);
            z -= (xi[0]-z)*r1*r2*r3;
        }
        else if (i == n-1)
        {
            r1=(1.0+0.235*bet)/(0.766+0.119*bet);
            r2=1.0/(1.0+0.639*(n-4.0)/(1.0+0.71*(n-4.0)));
            r3=1.0/(1.0+20.0*alf/((7.5+alf)*n*n));
            z += (z-xi[n-4])*r1*r2*r3;
        }
        else if (i == n)
        {
            r1=(1.0+0.37*bet)/(1.67+0.28*bet);
            r2=1.0/(1.0+0.22*(n-8.0)/n);
            r3=1.0/(1.0+8.0*alf/((6.28+alf)*n*n));
            z += (z-xi[n-3])*r1*r2*r3;
        }
        else
        {
            z=3.0*xi[i-2]-3.0*xi[i-3]+xi[i-4];
        }
        alfbet=alf+bet;
        its=1;
        do
        {
            temp=2.0+alfbet;
            p1=(alf-bet+temp*z)/2.0;
            p2=1.0;
            for (j=2; j<=n; j++)
            {
                p3=p2;
                p2=p1;
                temp=2*j+alfbet;
                a=2*j*(j+alfbet)*(temp-2.0);
                b=(temp-1.0)*(alf*alf-bet*bet+temp*(temp-2.0)*z);
                c=2.0*(j-1+alf)*(j-1+bet)*temp;
                p1=(b*p2-c*p3)/a;
            }
            pp=(n*(alf-bet-temp*z)*p1+2.0*(n+alf)*(n+bet)*p2)/(temp*(1.0-z*z));
            z1=z;
            z=z1-p1/pp;
            its++;
        } while(std::abs(z-z1) > EPS && its <= MAXIT);
        if (its > MAXIT)
            std::cout << "too many iterations in gaujac";
        xi[i-1]=z;
        if (bet == 0.0)
        {
            wi[i-1] = 1.0 / ( (n+alf) * n ) *temp* std::pow(2.0,alfbet)/(pp*p2);
        }
        else
        {
            wi[i-1]= std::exp(  std::lgamma(alf+n) + std::lgamma(bet+n) - std::lgamma(n+1.0)- std::lgamma(n+alfbet+1.0))*temp*pow(2.0,alfbet)/(pp*p2);
        }
    }
    for (int i = 0; i < n; i++)
    {
        wi[i] *= 0.5 * std::pow(1-xi[i],-alf) * std::pow(1+xi[i], -bet);
        xi[i] = 0.5 * (xi[i]+1);
    }
}


/** \brief Computes weights and integration points for the Gauss-Legendre Quadrature
           \f$ \int_{-1}^1 f(x) dx \f$.
           Transformed from [-1,1] to [0,1].
 *
 * \param[in] n Order of integration rule
 * \param[out] xi vector of integration points
 * \param[out] wi vector of integration weights
 * \return
 *
 */
template<typename T>
inline void computeGaussRule(const size_t n, std::vector<T> & xi, std::vector<T> & wi)
{
    xi.resize (n);
    wi.resize (n);
    // the roots are symmetric in the interval, so we only have to find half of
    // them
    size_t m = (n+1)/2;
    T p1, p2, p3;
    T pp, z, z1;
    constexpr T MYPI = std::acos(T(-1));
    constexpr T oneoverfour = static_cast<T>(T(1) / T(4));
    constexpr T oneovertwo  = static_cast<T>(T(1) / T(2));
    for (size_t i = 1; i <= m; i++)
    {
        z = std::cos(static_cast<T>(MYPI * (i - oneoverfour) / (n + oneovertwo)));
        do
        {
            p1 = 1;
            p2 = 0;
            for (size_t j = 1; j <= n; j++)
            {
                p3 = p2;
                p2 = p1;
                p1 = ((2 * j - 1) * z * p2 - (j - 1) * p3) / j;
            }
            // p1 is legendre polynomial
            pp = n * (z*p1-p2) / (z*z - 1);
            z1 = z;
            z = z1-p1/pp;
        }
        while(std::abs(z - z1) > T(1e-17));
        xi[i-1] = oneovertwo * (1 - z);
        xi[n-i] = oneovertwo * (1 + z);
        wi[i-1] = wi[n-i] = T(1) / ( (1  - z * z) * pp * pp);
    }
}
